// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#ifndef BAUMWELCH_DECODE_H_
#define BAUMWELCH_DECODE_H_

// This header defines functions for decoding ciphertext into plaintext using
// the Viterbi algorithm. Both FST and FAR inputs/outputs are supported.
//
// If the semiring (arc) does not have the path property, then there must exist
// a weight converter to and from the tropical semiring (arc). (Weight
// properties must also be known at compile time.)
//
// Decoding proceeds as follows:
//
// 1. Compose the channel model and the ciphertext.
// 2. Project onto the input (plaintext) side of (1).
// 3. Compose the plaintext with the result of (2) to construct a weighted
//    lattice.
// 4. If the input semiring is does not have the path property (i.e., is not
//    kPath), attempt to convert the weighted lattice to the tropical semiring.
// 5. Compute the shortest path over the weighted lattice (3-4).
// 6. If the input semiring does not have the path property (i.e., is not
//    kPath), attempt to convert the shortest path (5) back to the input
//    semiring.
// 7. Project onto the input, remove epsilon arcs, and remove weights.
//
// Note that if a uniform channel model is provided, this is equivalent to
// decoding with the plaintext model alone.
//
// By convention, operations in this library that work with FarReader input
// reset the FAR to its initial position upon completion.

#include <algorithm>
#include <memory>
#include <numeric>
#include <type_traits>
#include <vector>

#include <fst/extensions/far/far.h>
#include <fst/fstlib.h>

#include <baumwelch/cascade.h>
#include <baumwelch/data.h>

namespace fst {
namespace internal {

// Power-mapper that takes a scalar power argument. This template is only valid
// for unsigned float weight types.
template <class A>
class FloatPowerMapper {
 public:
  using FromArc = A;
  using ToArc = A;
  using Weight = typename FromArc::Weight;

  explicit FloatPowerMapper(float power) : power_(power) {}

  ToArc operator()(const FromArc &arc) const {
    return ToArc(arc.ilabel, arc.olabel, Power(arc.weight, power_),
                 arc.nextstate);
  }

  constexpr MapFinalAction FinalAction() const { return MAP_NO_SUPERFINAL; }

  constexpr MapSymbolsAction InputSymbolsAction() const {
    return MAP_COPY_SYMBOLS;
  }

  constexpr MapSymbolsAction OutputSymbolsAction() const {
    return MAP_COPY_SYMBOLS;
  }

  uint64 Properties(uint64 props) const {
    return props & kWeightInvariantProperties;
  }

 private:
  const float power_;
};

// Helper function that reverses the state numbering of the output of single
// ShortestPath so that the FST is kString.
template <class Arc>
void StateReverse(MutableFst<Arc> *fst) {
  using StateId = typename Arc::StateId;
  std::vector<StateId> order(fst->NumStates());
  std::iota(order.rbegin(), order.rend(), 0);
  StateSort(fst, order);
}

// Computes the shortest path over a semiring with the path property:
//
// * Compute the shortest path using Viterbi search
// * Build epsilon-free WFSA
// * Optimize to yield a kString FST
template <class Arc,
          typename std::enable_if<(Arc::Weight::Properties() & kPath) ==
                                  kPath>::type * = nullptr>
void ComputeShortestPath(const ConstFst<Arc> &ico, VectorFst<Arc> *plaintext) {
  ShortestPath(ico, plaintext);
  // Convert to epsilon-free WFSA.
  Project(plaintext, PROJECT_INPUT);
  RmEpsilon(plaintext);
  // Optimize to a kString FST.
  ArcMap(plaintext, RmWeightMapper<Arc>());
  StateReverse(plaintext);
}

// Computes the number of explored states in a distance vector. This is used to
// log the expansion of the DFA in A* search.
template <class Weight>
size_t CountExploredStates(const std::vector<Weight> &distance) {
  static const auto nonzero_weight = [](const Weight &weight) {
    return weight != Weight::Zero();
  };
  return std::count_if(distance.begin(), distance.end(), nonzero_weight);
}

// Computes the shortest path over a semiring without the path property:
//
// * Build epsilon-free WFSA
// * Lazily determinize and lazily compute the A* heuristic
// * Lazily cast into the tropical semiring
// * Compute the shortest path using A* search
// * Eagerly cast back into the input semiring
// * Optimize to yield a kString FST.
template <class Arc,
          typename std::enable_if<(Arc::Weight::Properties() & kPath) !=
                                  kPath>::type * = nullptr>
void ComputeShortestPath(const ConstFst<Arc> &ico, VectorFst<Arc> *plaintext) {
  // Build epsilon-free WFSA.
  VectorFst<Arc> lattice;
  Project(ico, &lattice, PROJECT_INPUT);
  RmEpsilon(&lattice);
  // A* search heuristic.
  using Weight = typename Arc::Weight;
  std::vector<Weight> nfa_beta;
  ShortestDistance(lattice, &nfa_beta, /*reverse=*/true);
  std::vector<Weight> dfa_beta;
  const DeterminizeFstOptions<Arc> opts;
  const DeterminizeFst<Arc> dfa(lattice, &nfa_beta, &dfa_beta, opts);
  // Computing shortest path.
  using StateId = typename Arc::StateId;
  using Weight = typename Arc::Weight;
  using PathWeight = TropicalWeightTpl<typename Weight::ValueType>;
  using PathArc = ArcTpl<PathWeight>;
  using ToPathMapper = WeightConvertMapper<Arc, PathArc>;
  using FromPathMapper = WeightConvertMapper<PathArc, Arc>;
  const ArcMapFst<Arc, PathArc, ToPathMapper> path_dfa(dfa, ToPathMapper());
  VectorFst<PathArc> shortest_path;
  {
    using MyEstimate = NaturalAStarEstimate<StateId, PathWeight>;
    using MyQueue = NaturalAStarQueue<StateId, PathWeight, MyEstimate>;
    using MyArcFilter = AnyArcFilter<PathArc>;
    using MyShortestPathOptions =
        ShortestPathOptions<PathArc, MyQueue, MyArcFilter>;
    const MyEstimate estimate(
        reinterpret_cast<const std::vector<PathWeight> &>(dfa_beta));
    std::vector<PathWeight> dfa_alpha;
    MyQueue queue(dfa_alpha, estimate);
    const MyArcFilter filter;
    const MyShortestPathOptions opts(
        &queue, filter,
        /*nshortest=*/1,           // 1-best.
        /*unique=*/false,          // Lattice is already deterministic.
        /*has_distance=*/true,     // Heuristic is precomputed.
        /*delta=*/kShortestDelta,  // Default.
        /*first_path=*/true);      // Heuristic is admissible.
    ShortestPath(path_dfa, &shortest_path, &dfa_alpha, opts);
    VLOG(1) << CountExploredStates(dfa_alpha) << " states explored";
  }
  ArcMap(shortest_path, plaintext, FromPathMapper());
  // Optimize to a kString FST.
  ArcMap(plaintext, RmWeightMapper<Arc>());
  StateReverse(plaintext);
}

template <class Arc, class Data>
void DecodeBaumWelch(Data *data, const Fst<Arc> &channel,
                     FarWriter<Arc> *hypotext) {
  using Weight = typename Arc::Weight;
  for (; !data->Done(); data->Next()) {
    const SingleStateCascade<Arc> cascade(data->GetPlaintext(),
                                          data->GetCiphertext(),
                                          channel);
    const ConstFst<Arc> ico(cascade.GetFst());
    VectorFst<Arc> plaintext;
    ComputeShortestPath(ico, &plaintext);
    hypotext->Add(data->GetKey(), CompactStringFst<Arc>(plaintext));
  }
  data->Reset();
}

}  // namespace internal

// This instantiates the paired construction.
template <class Arc>
void DecodeBaumWelch(FarReader<Arc> *plaintext,
                     FarReader<Arc> *ciphertext,
                     const Fst<Arc> &channel,
                     FarWriter<Arc> *writer) {
  internal::PairedData<Arc> data(plaintext, ciphertext);
  internal::DecodeBaumWelch(&data, channel, writer);
}

// This instantiates the decipherment construction; note that if the sparsity
// penalty is defined, the likelihood will be degenerate.
template <class Arc>
void DecodeBaumWelch(const Fst<Arc> &plaintext,
                     FarReader<Arc> *ciphertext,
                     const Fst<Arc> &channel,
                     FarWriter<Arc> *writer,
                     float power = 1.0) {
  internal::DeciphermentData<Arc> data(plaintext, ciphertext);
  if (power > 1.0) {
    VectorFst<Arc> sparse_channel;
    ArcMap(channel, &sparse_channel, internal::FloatPowerMapper<Arc>(power));
    internal::DecodeBaumWelch(&data, sparse_channel, writer);
  } else {
    internal::DecodeBaumWelch(&data, channel, writer);
  }
}

}  // namespace fst

#endif  // BAUMWELCH_DECODE_H_

